import datetime as dt
from unittest.mock import patch

from django.utils.dateparse import parse_datetime
from django.utils.timezone import now
from esi.errors import TokenError, TokenExpiredError
from esi.models import Token
from eveuniverse.models import EveType

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import NoSocketsTestCase

from blueprints.models import Blueprint, Location, Owner, Request

from . import add_character_to_user, create_owner, create_user_from_evecharacter
from .testdata.esi_client_stub import esi_client_stub
from .testdata.factory import BlueprintFactory, LocationStationFactory, OwnerFactory
from .testdata.load_entities import load_entities
from .testdata.load_eveuniverse import load_eveuniverse
from .testdata.load_locations import load_locations

MANAGERS_PATH = "blueprints.managers"
MODELS_PATH = "blueprints.models"


class TestBlueprintsBase(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_entities()
        load_eveuniverse()
        load_locations()


class TestBlueprintQuerySet(TestBlueprintsBase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        # given
        cls.owner_1001 = create_owner(character_id=1001, corporation_id=None)
        Blueprint.objects.create(
            location=Location.objects.get(id=60003760),
            eve_type=EveType.objects.get(id=33519),
            owner=cls.owner_1001,
            runs=10,
            location_flag="AssetSafety",
            material_efficiency=10,
            time_efficiency=30,
            item_id=1,
        )
        cls.owner_1002 = create_owner(character_id=1002, corporation_id=2001)
        Blueprint.objects.create(
            location=Location.objects.get(id=1000000000001),
            eve_type=EveType.objects.get(id=33519),
            owner=cls.owner_1002,
            location_flag="AssetSafety",
            material_efficiency=20,
            time_efficiency=40,
            item_id=2,
        )

    def test_should_annotate_is_bpo(self):
        # when
        result = Blueprint.objects.all().annotate_is_bpo().values()
        # then
        obj = result[0]
        self.assertEqual(obj["item_id"], 1)
        self.assertEqual(obj["is_bpo"], "no")
        obj = result[1]
        self.assertEqual(obj["item_id"], 2)
        self.assertEqual(obj["is_bpo"], "yes")

    def test_should_annotate_owner_name(self):
        # when
        result = Blueprint.objects.all().annotate_owner_name().values()
        # then
        obj = result[0]
        self.assertEqual(obj["item_id"], 1)
        self.assertEqual(obj["owner_name"], "Bruce Wayne")
        obj = result[1]
        self.assertEqual(obj["item_id"], 2)
        self.assertEqual(obj["owner_name"], "Wayne Technologies")


class TestBlueprintManagerUserHasAccess(TestBlueprintsBase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        # given
        cls.owner_1002 = create_owner(character_id=1002, corporation_id=2001)
        Blueprint.objects.create(
            location=Location.objects.get(id=60003760),
            eve_type=EveType.objects.get(id=33519),
            owner=cls.owner_1002,
            location_flag="AssetSafety",
            material_efficiency=10,
            time_efficiency=30,
            item_id=2,
        )
        cls.owner_1003 = create_owner(character_id=1003, corporation_id=2002)
        Blueprint.objects.create(
            location=Location.objects.get(id=60003760),
            eve_type=EveType.objects.get(id=33519),
            owner=cls.owner_1003,
            location_flag="AssetSafety",
            material_efficiency=10,
            time_efficiency=30,
            item_id=3,
        )
        cls.owner_1101 = create_owner(character_id=1101, corporation_id=2101)
        Blueprint.objects.create(
            location=Location.objects.get(id=60003760),
            eve_type=EveType.objects.get(id=33519),
            owner=cls.owner_1101,
            location_flag="AssetSafety",
            material_efficiency=10,
            time_efficiency=30,
            item_id=4,
        )
        cls.owner_1102 = create_owner(character_id=1102, corporation_id=2102)
        Blueprint.objects.create(
            location=Location.objects.get(id=60003760),
            eve_type=EveType.objects.get(id=33519),
            owner=cls.owner_1102,
            location_flag="AssetSafety",
            material_efficiency=10,
            time_efficiency=30,
            item_id=5,
        )
        cls.owner_1004 = create_owner(character_id=1004, corporation_id=None)
        Blueprint.objects.create(
            location=Location.objects.get(id=60003760),
            eve_type=EveType.objects.get(id=33519),
            owner=cls.owner_1004,
            location_flag="AssetSafety",
            material_efficiency=10,
            time_efficiency=30,
            item_id=6,
        )

    def setUp(self) -> None:
        # given
        self.owner_1001 = create_owner(character_id=1001, corporation_id=None)
        self.user = self.owner_1001.character.user
        add_character_to_user(self.user, EveCharacter.objects.get(character_id=1103))
        Blueprint.objects.create(
            location=Location.objects.get(id=60003760),
            eve_type=EveType.objects.get(id=33519),
            owner=self.owner_1001,
            location_flag="AssetSafety",
            material_efficiency=10,
            time_efficiency=30,
            item_id=1,
        )

    def test_should_return_personal_and_corporation_and_alt_corporation(self):
        # when
        qs = Blueprint.objects.user_has_access(self.user)
        # then
        self.assertSetEqual(set(qs.values_list("item_id", flat=True)), {1, 2, 5})

    def test_should_return_personal_and_corporation_and_alt_corporation_and_alliance(
        self,
    ):
        # given
        self.user = AuthUtils.add_permission_to_user_by_name(
            "blueprints.view_alliance_blueprints", self.user
        )
        # when
        qs = Blueprint.objects.user_has_access(self.user)
        # then
        self.assertSetEqual(set(qs.values_list("item_id", flat=True)), {1, 2, 3, 5, 6})


class TestBlueprintManagerAnnotateLocationName(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.owner = OwnerFactory()

    def test_should_return_name(self):
        # given
        location = LocationStationFactory(name="Alpha")
        BlueprintFactory(owner=self.owner, location=location)
        # when
        qs = Blueprint.objects.annotate_location_name()
        # then
        obj = qs.first()
        self.assertEqual(obj.location_name, "Alpha")

    def test_should_return_parent_name(self):
        # given
        parent_location = LocationStationFactory(name="Parent")
        child_location = LocationStationFactory(name="", parent=parent_location)
        BlueprintFactory(owner=self.owner, location=child_location)
        # when
        qs = Blueprint.objects.annotate_location_name()
        # then
        obj = qs.first()
        self.assertEqual(obj.location_name, "Parent")

    def test_should_return_generic_name(self):
        # given
        location = LocationStationFactory(name="")
        BlueprintFactory(owner=self.owner, location=location)
        # when
        qs = Blueprint.objects.annotate_location_name()
        # then
        obj = qs.first()
        self.assertEqual(obj.location_name, f"Location #{location.id}")

    def test_should_return_parent_parent_name(self):
        # given
        parent_location = LocationStationFactory(name="Parent")
        child_location = LocationStationFactory(name="", parent=parent_location)
        child_child_location = LocationStationFactory(name="", parent=child_location)
        BlueprintFactory(owner=self.owner, location=child_child_location)
        # when
        qs = Blueprint.objects.annotate_location_name()
        # then
        obj = qs.first()
        self.assertEqual(obj.location_name, "Parent")

    def test_should_return_parent_parent_parent_name(self):
        # given
        parent_location = LocationStationFactory(name="Parent")
        child_location = LocationStationFactory(name="", parent=parent_location)
        child_child_location = LocationStationFactory(name="", parent=child_location)
        child_child_child_location = LocationStationFactory(
            name="", parent=child_child_location
        )
        BlueprintFactory(owner=self.owner, location=child_child_child_location)
        # when
        qs = Blueprint.objects.annotate_location_name()
        # then
        obj = qs.first()
        self.assertEqual(obj.location_name, "Parent")

    def test_should_return_first_parent_with_name_with_4_nodes(self):
        # given
        parent_location = LocationStationFactory(name="Parent")
        child_location = LocationStationFactory(name="Child", parent=parent_location)
        child_child_location = LocationStationFactory(name="", parent=child_location)
        child_child_child_location = LocationStationFactory(
            name="", parent=child_child_location
        )
        BlueprintFactory(owner=self.owner, location=child_child_child_location)
        # when
        qs = Blueprint.objects.annotate_location_name()
        # then
        obj = qs.first()
        self.assertEqual(obj.location_name, "Child")

    def test_should_return_first_parent_with_name_with_3_nodes(self):
        # given
        parent_location = LocationStationFactory(name="Parent")
        child_location = LocationStationFactory(name="Child", parent=parent_location)
        child_child_location = LocationStationFactory(name="", parent=child_location)
        BlueprintFactory(owner=self.owner, location=child_child_location)
        # when
        qs = Blueprint.objects.annotate_location_name()
        # then
        obj = qs.first()
        self.assertEqual(obj.location_name, "Child")


@patch(MODELS_PATH + ".esi")
@patch(MANAGERS_PATH + ".esi")
class TestCorporateOwner(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_entities()
        load_eveuniverse()
        load_locations()

    def setUp(self) -> None:
        self.owner = create_owner(character_id=1101, corporation_id=2101)

    def test_should_return_corporation_name_for_owner(
        self, mock_esi_managers, mock_esi_models
    ):
        # when
        result = str(self.owner)
        # then
        self.assertEqual(result, "Lexcorp")

    def test_should_return_corporation_name_for_owner_with_no_character(
        self, mock_esi_managers, mock_esi_models
    ):
        # given
        owner = Owner.objects.create(
            corporation=EveCorporationInfo.objects.get(corporation_id=2001)
        )
        # when
        result = str(owner)
        # then
        self.assertEqual(result, "Wayne Technologies")

    def test_should_return_empty_string_for_empty_owner(
        self, mock_esi_managers, mock_esi_models
    ):
        # given
        owner = Owner.objects.create()
        # when
        result = str(owner)
        # then
        self.assertEqual(result, "")

    def test_update_locations_esi(self, mock_esi_managers, mock_esi_models):
        mock_esi_managers.client = esi_client_stub
        mock_esi_models.client = esi_client_stub

        self.owner.update_locations_esi()

        self.assertEquals(
            Location.objects.get(id=1100000000001).parent,
            Location.objects.get(id=60003760),
        )

    def test_update_blueprints_esi(self, mock_esi_managers, mock_esi_models):
        mock_esi_managers.client = esi_client_stub
        mock_esi_models.client = esi_client_stub
        self.owner.update_blueprints_esi()
        self.assertEquals(Blueprint.objects.filter(eve_type_id=33519).count(), 1)

    def test_should_update_industry_jobs_esi(self, mock_esi_managers, mock_esi_models):
        # given

        mock_esi_managers.client = esi_client_stub
        mock_esi_models.client = esi_client_stub
        self.owner.update_blueprints_esi()
        # when
        self.owner.update_industry_jobs_esi()
        # then
        self.assertEquals(self.owner.jobs.count(), 1)
        obj = self.owner.jobs.first()
        self.assertEqual(obj.id, 100000002)
        self.assertEqual(obj.activity, 5)
        self.assertEqual(obj.location_id, 1000000000001)
        self.assertEqual(obj.installer, EveCharacter.objects.get(character_id=1001))
        self.assertEqual(obj.runs, 1)
        self.assertEqual(obj.start_date, parse_datetime("2020-12-21T23:37:14Z"))
        self.assertEqual(obj.status, "active")


class TestLocationNamePlus(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_should_return_name(self):
        # given
        location = LocationStationFactory(name="Alpha")
        # when/then
        self.assertEqual(location.full_qualified_name(), "Alpha")

    def test_should_return_parent_name(self):
        # given
        parent_location = LocationStationFactory(name="Parent")
        location = LocationStationFactory(name="Child", parent=parent_location)
        # when/then
        self.assertEqual(location.full_qualified_name(), "Parent - Child")

    def test_should_return_generic_name(self):
        # given
        location = LocationStationFactory(name="")
        # when/then
        self.assertEqual(location.full_qualified_name(), f"Location #{location.id}")

    def test_should_return_parent_parent_parent_name(self):
        # given
        parent_location = LocationStationFactory(name="Parent")
        child_location = LocationStationFactory(name="Child 1", parent=parent_location)
        child_child_location = LocationStationFactory(
            name="Child 2", parent=child_location
        )
        location = LocationStationFactory(name="Child 3", parent=child_child_location)
        # when/then
        self.assertEqual(
            location.full_qualified_name(), "Parent - Child 1 - Child 2 - Child 3"
        )


class TestOwnerValidToken(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.owner = OwnerFactory()

    def test_should_return_valid_token(self):
        # when
        result = self.owner.valid_token(["esi-characters.read_blueprints.v1"])
        # then
        self.assertIsInstance(result, Token)

    def test_should_raise_error_when_no_token_with_requested_scope_found(self):
        # when/then
        with self.assertRaises(TokenError):
            self.owner.valid_token(["unknown-scope"])

    @patch(MODELS_PATH + ".Token.objects.filter")
    def test_should_raise_error_when_token_has_issue(self, mock):
        # given
        mock.side_effect = TokenExpiredError
        # when/then
        with self.assertRaises(TokenExpiredError):
            self.owner.valid_token(["esi-characters.read_blueprints.v1"])


@patch(MODELS_PATH + ".esi")
@patch(MANAGERS_PATH + ".esi")
class TestPersonalOwner(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_entities()
        load_eveuniverse()
        load_locations()

    def setUp(self) -> None:
        self.owner = create_owner(character_id=1101, corporation_id=None)

    def test_update_locations_esi(self, mock_esi_managers, mock_esi_models):
        mock_esi_managers.client = esi_client_stub
        mock_esi_models.client = esi_client_stub

        self.owner.update_locations_esi()

        self.assertEquals(
            Location.objects.get(id=1100000000001).parent,
            Location.objects.get(id=60003760),
        )

    def test_update_blueprints_esi(self, mock_esi_managers, mock_esi_models):
        mock_esi_managers.client = esi_client_stub
        mock_esi_models.client = esi_client_stub
        self.owner.update_blueprints_esi()
        self.assertEquals(Blueprint.objects.filter(eve_type_id=33519).count(), 1)

    def test_should_update_industry_jobs_esi(self, mock_esi_managers, mock_esi_models):
        # given

        mock_esi_managers.client = esi_client_stub
        mock_esi_models.client = esi_client_stub
        self.owner.update_blueprints_esi()
        # when
        self.owner.update_industry_jobs_esi()
        # then
        self.assertEquals(self.owner.jobs.count(), 1)
        obj = self.owner.jobs.first()
        self.assertEqual(obj.id, 100000001)
        self.assertEqual(obj.activity, 5)
        self.assertEqual(obj.location_id, 1000000000001)
        self.assertEqual(obj.installer, EveCharacter.objects.get(character_id=1001))
        self.assertEqual(obj.runs, 1)
        self.assertEqual(obj.start_date, parse_datetime("2020-12-21T23:37:14Z"))
        self.assertEqual(obj.status, "active")


class TestRequests(TestBlueprintsBase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        # given
        cls.user_1002, _ = create_user_from_evecharacter(1002)
        cls.owner_1001 = create_owner(character_id=1001, corporation_id=2001)
        cls.user_1001 = cls.owner_1001.character.user
        cls.bp_1 = Blueprint.objects.create(
            location=Location.objects.get(id=60003760),
            eve_type=EveType.objects.get(id=33519),
            owner=cls.owner_1001,
            location_flag="AssetSafety",
            material_efficiency=10,
            time_efficiency=30,
            item_id=2,
        )
        cls.owner_1101 = create_owner(character_id=1101, corporation_id=2101)
        cls.bp_2 = Blueprint.objects.create(
            location=Location.objects.get(id=60003760),
            eve_type=EveType.objects.get(id=33519),
            owner=cls.owner_1101,
            location_flag="AssetSafety",
            material_efficiency=10,
            time_efficiency=30,
            item_id=3,
        )

    def test_should_strings(self):
        # given
        req_1 = Request.objects.create(
            blueprint=self.bp_1,
            requesting_user=self.user_1002,
            status=Request.STATUS_OPEN,
        )
        # when/then
        self.assertTrue(str(req_1))
        self.assertTrue(repr(req_1))

    def test_should_return_fulfillable_requests(self):
        # given
        req_1 = Request.objects.create(
            blueprint=self.bp_1,
            requesting_user=self.user_1002,
            status=Request.STATUS_OPEN,
        )
        Request.objects.create(
            blueprint=self.bp_1,
            requesting_user=self.user_1002,
            status=Request.STATUS_FULFILLED,
        )
        Request.objects.create(
            blueprint=self.bp_2,
            requesting_user=self.user_1002,
            status=Request.STATUS_OPEN,
        )
        Request.objects.create(
            blueprint=self.bp_1,
            requesting_user=self.user_1002,
            status=Request.STATUS_OPEN,
            closed_at=now() - dt.timedelta(days=1),
        )
        # when
        result = Request.objects.all().requests_fulfillable_by_user(self.user_1001)
        # then
        result_pks = set(result.values_list("pk", flat=True))
        self.assertSetEqual(result_pks, {req_1.pk})

    def test_should_return_requests_being_fulfilled(self):
        # given
        req_1 = Request.objects.create(
            blueprint=self.bp_1,
            requesting_user=self.user_1002,
            status=Request.STATUS_IN_PROGRESS,
            fulfulling_user=self.user_1001,
        )
        Request.objects.create(
            blueprint=self.bp_1,
            requesting_user=self.user_1002,
            status=Request.STATUS_OPEN,
        )
        Request.objects.create(
            blueprint=self.bp_1,
            requesting_user=self.user_1002,
            status=Request.STATUS_FULFILLED,
        )
        Request.objects.create(
            blueprint=self.bp_2,
            requesting_user=self.user_1002,
            status=Request.STATUS_OPEN,
        )
        Request.objects.create(
            blueprint=self.bp_1,
            requesting_user=self.user_1002,
            status=Request.STATUS_IN_PROGRESS,
            fulfulling_user=self.user_1001,
            closed_at=now() - dt.timedelta(days=1),
        )
        # when
        result = Request.objects.all().requests_being_fulfilled_by_user(self.user_1001)
        # then
        result_pks = set(result.values_list("pk", flat=True))
        self.assertSetEqual(result_pks, {req_1.pk})

    def test_should_return_open_requests_total_count(self):
        # given
        Request.objects.create(
            blueprint=self.bp_1,
            requesting_user=self.user_1002,
            status=Request.STATUS_IN_PROGRESS,
            fulfulling_user=self.user_1001,
        )
        Request.objects.create(
            blueprint=self.bp_1,
            requesting_user=self.user_1002,
            status=Request.STATUS_OPEN,
        )
        Request.objects.create(
            blueprint=self.bp_1,
            requesting_user=self.user_1002,
            status=Request.STATUS_FULFILLED,
        )
        Request.objects.create(
            blueprint=self.bp_2,
            requesting_user=self.user_1002,
            status=Request.STATUS_OPEN,
        )
        Request.objects.create(
            blueprint=self.bp_1,
            requesting_user=self.user_1002,
            status=Request.STATUS_IN_PROGRESS,
            fulfulling_user=self.user_1001,
            closed_at=now() - dt.timedelta(days=1),
        )
        # when
        result = Request.objects.open_requests_total_count(self.user_1001)
        # then
        self.assertEqual(result, 2)



from django.core.exceptions import ValidationError
from django.test import TestCase

from blueprints.validators import validate_material_efficiency, validate_time_efficiency


class TestValidateMaterialEfficiencyValue(TestCase):
    def test_should_pass_when_valid(self):
        for value in [0, 4, 10]:
            with self.subTest(value=value):
                self.assertIsNone(validate_material_efficiency(value))

    def test_should_raise_error_when_invalid(self):
        for value in [-5, -0.1, 10.1, 20]:
            with self.subTest(value=value):
                with self.assertRaises(ValidationError):
                    validate_material_efficiency(value)


class TestValidateTimeEfficiencyValue(TestCase):
    def test_should_pass_when_valid(self):
        for value in [0, 4, 10, 20]:
            with self.subTest(value=value):
                self.assertIsNone(validate_time_efficiency(value))

    def test_should_raise_error_when_invalid(self):
        for value in [-5, -0.1, 3, 7, 20.1, 100]:
            with self.subTest(value=value):
                with self.assertRaises(ValidationError):
                    validate_time_efficiency(value)



import random
import socket
import string
from datetime import datetime, timedelta

from django.contrib.auth.models import User
from django.test import TestCase
from esi.models import Scope, Token

from allianceauth.eveonline.models import EveCharacter


def _get_random_string(char_count):
    return "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(char_count)
    )


def _dt_eveformat(dt: object) -> str:
    """converts a datetime to a string in eve format
    e.g. '2019-06-25T19:04:44'
    """
    dt2 = datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second)
    return dt2.isoformat()


def _generate_token(
    character_id: int,
    character_name: str,
    access_token: str = "access_token",
    refresh_token: str = "refresh_token",
    scopes: list = None,
    timestamp_dt: object = None,
    expires_in: int = 1200,
) -> dict:
    if timestamp_dt is None:
        timestamp_dt = datetime.utcnow()
    if scopes is None:
        scopes = [
            "esi-universe.read_structures.v1",
            "esi-corporations.read_blueprints.v1",
            "esi-assets.read_corporation_assets.v1",
            "esi-characters.read_blueprints.v1",
            "esi-assets.read_assets.v1",
            "esi-industry.read_character_jobs.v1",
            "esi-industry.read_corporation_jobs.v1",
        ]
    token = {
        "access_token": access_token,
        "token_type": "Bearer",
        "expires_in": expires_in,
        "refresh_token": refresh_token,
        "timestamp": int(timestamp_dt.timestamp()),
        "CharacterID": character_id,
        "CharacterName": character_name,
        "ExpiresOn": _dt_eveformat(timestamp_dt + timedelta(seconds=expires_in)),
        "Scopes": " ".join(list(scopes)),
        "TokenType": "Character",
        "CharacterOwnerHash": _get_random_string(28),
        "IntellectualProperty": "EVE",
    }
    return token


def _store_as_Token(token: dict, user: object) -> object:
    """Stores a generated token dict as Token object for given user

    returns Token object
    """
    obj = Token.objects.create(
        access_token=token["access_token"],
        refresh_token=token["refresh_token"],
        user=user,
        character_id=token["CharacterID"],
        character_name=token["CharacterName"],
        token_type=token["TokenType"],
        character_owner_hash=token["CharacterOwnerHash"],
    )
    for scope_name in token["Scopes"].split(" "):
        scope, _ = Scope.objects.get_or_create(name=scope_name)
        obj.scopes.add(scope)

    return obj


def add_new_token(user: User, character: EveCharacter, scopes: list) -> Token:
    """generates a new Token for the given character and adds it to the user"""
    return _store_as_Token(
        _generate_token(
            character_id=character.character_id,
            character_name=character.character_name,
            scopes=scopes,
        ),
        user,
    )


class SocketAccessError(Exception):
    pass


class NoSocketsTestCase(TestCase):
    """Variation of TestCase class that prevents any use of sockets"""

    @classmethod
    def setUpClass(cls):
        cls.socket_original = socket.socket
        socket.socket = cls.guard
        return super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        socket.socket = cls.socket_original
        return super().tearDownClass()

    @staticmethod
    def guard(*args, **kwargs):
        raise SocketAccessError("Attempted to access network")



# Generated by Django 3.1.2 on 2021-02-01 14:40

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("blueprints", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="general",
            options={
                "default_permissions": (),
                "managed": False,
                "permissions": (
                    ("basic_access", "Can access this app"),
                    ("request_blueprints", "Can request blueprints"),
                    ("manage_requests", "Can review and accept blueprint requests"),
                    (
                        "add_personal_blueprint_owner",
                        "Can add personal blueprint owners",
                    ),
                    (
                        "add_corporate_blueprint_owner",
                        "Can add corporate blueprint owners",
                    ),
                    ("view_alliance_blueprints", "Can view alliance's blueprints"),
                    (
                        "view_blueprint_locations",
                        "Can view the location of all blueprints",
                    ),
                    (
                        "view_industry_jobs",
                        "Can view details about running industry jobs",
                    ),
                ),
            },
        ),
    ]



"""Tasks for Blueprints."""

import random

from bravado.exception import HTTPBadGateway, HTTPGatewayTimeout, HTTPServiceUnavailable
from celery import shared_task

from esi.models import Token

from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce
from app_utils.esi import EsiErrorLimitExceeded, EsiOffline
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import BLUEPRINTS_TASKS_TIME_LIMIT
from .models import Location, Owner

DEFAULT_TASK_PRIORITY = 6


logger = LoggerAddTag(get_extension_logger(__name__), __title__)

TASK_DEFAULT_KWARGS = {
    "time_limit": BLUEPRINTS_TASKS_TIME_LIMIT,
}

TASK_ESI_KWARGS = {
    **TASK_DEFAULT_KWARGS,
    **{
        "autoretry_for": (
            OSError,
            HTTPBadGateway,
            HTTPGatewayTimeout,
            HTTPServiceUnavailable,
        ),
        "retry_kwargs": {"max_retries": 3},
        "retry_backoff": 30,
    },
}


@shared_task(
    **{
        **TASK_ESI_KWARGS,
        **{
            "base": QueueOnce,
            "once": {"keys": ["owner_pk"], "graceful": True},
            "max_retries": None,
        },
    }
)
def update_blueprints_for_owner(owner_pk: int):
    """Fetch all blueprints for an owner from ESI."""
    owner = Owner.objects.get(pk=owner_pk)
    owner.update_blueprints_esi()


@shared_task(
    **{
        **TASK_ESI_KWARGS,
        **{
            "base": QueueOnce,
            "once": {"keys": ["owner_pk"], "graceful": True},
            "max_retries": None,
        },
    }
)
def update_industry_jobs_for_owner(owner_pk: int):
    """Fetch all industry jobs for an owner from ESI."""
    owner = Owner.objects.get(pk=owner_pk)
    owner.update_industry_jobs_esi()


@shared_task(
    **{
        **TASK_ESI_KWARGS,
        **{
            "base": QueueOnce,
            "once": {"keys": ["owner_pk"], "graceful": True},
            "max_retries": None,
        },
    }
)
def update_locations_for_owner(owner_pk: int):
    """Fetch all blueprints for an owner from ESI."""
    owner = Owner.objects.get(pk=owner_pk)
    owner.update_locations_esi()


@shared_task(**TASK_DEFAULT_KWARGS)
def update_all_blueprints():
    """Update all blueprints."""
    for owner in Owner.objects.filter(is_active=True):
        update_blueprints_for_owner.apply_async(
            kwargs={"owner_pk": owner.pk}, priority=DEFAULT_TASK_PRIORITY
        )


@shared_task(**TASK_DEFAULT_KWARGS)
def update_all_industry_jobs():
    """Update all industry jobs."""
    for owner in Owner.objects.filter(is_active=True):
        update_industry_jobs_for_owner.apply_async(
            kwargs={"owner_pk": owner.pk}, priority=DEFAULT_TASK_PRIORITY
        )


@shared_task(**TASK_DEFAULT_KWARGS)
def update_all_locations():
    """Update all locations."""
    for owner in Owner.objects.filter(is_active=True):
        update_locations_for_owner.apply_async(
            kwargs={"owner_pk": owner.pk}, priority=DEFAULT_TASK_PRIORITY
        )


@shared_task(
    **{
        **TASK_ESI_KWARGS,
        **{
            "bind": True,
            "base": QueueOnce,
            "once": {"keys": ["id"], "graceful": True},
            "max_retries": None,
        },
    }
)
def update_structure_esi(self, id: int, token_pk: int):
    """Updates a structure object from ESI
    and retries later if the ESI error limit has already been reached
    """
    token = Token.objects.get(pk=token_pk)

    try:
        Location.objects.structure_update_or_create_esi(id=id, token=token)

    except EsiOffline as ex:
        logger.warning(
            "Location #%s: ESI appears to be offline. Trying again in 30 minutes.", id
        )
        raise self.retry(countdown=30 * 60 + int(random.uniform(1, 20))) from ex

    except EsiErrorLimitExceeded as ex:
        logger.warning(
            "Location #%s: ESI error limit threshold reached. "
            "Trying again in %s seconds",
            id,
            ex.retry_in,
        )
        raise self.retry(countdown=ex.retry_in) from ex



from django.test import TestCase
from eveuniverse.tools.testdata import ModelSpec, create_testdata

from . import eveuniverse_test_data_filename


class CreateEveUniverseTestData(TestCase):
    def test_create_testdata(self):
        testdata_spec = [
            ModelSpec("EveSolarSystem", ids=[30000142, 30004984, 30001161, 30002537]),
            ModelSpec(
                "EveType",
                ids=[
                    5,
                    23,
                    603,
                    20185,
                    24311,
                    24312,
                    35832,
                    35835,
                    52678,
                ],
            ),
            ModelSpec(
                "EveType",
                ids=[
                    19540,
                    19551,
                    19553,
                    33519,
                ],
                include_children=True,
            ),
        ]
        create_testdata(testdata_spec, eveuniverse_test_data_filename())



# flake8: noqa
"""scripts generates large amount of blueprints for load testing

This script can be executed directly from shell.
Please make sure to set your user ID via environment variable: BLUEPRINTS_USER_ID
"""

import inspect
import os
import sys

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
myauth_dir = (
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(currentdir))))
    + "/myauth"
)
sys.path.insert(0, myauth_dir)

import django
from django.apps import apps
from django.db import transaction
from django.utils.timezone import now

# init and setup django project
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

if not apps.is_installed("blueprints"):
    raise RuntimeError("The app blueprints is not installed")

"""SCRIPT"""
from random import randint

from django.contrib.auth.models import User
from eveuniverse.models import EveSolarSystem, EveType

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.tests.auth_utils import AuthUtils

from blueprints.models import Blueprint, EveSolarSystem, EveType, Location, Owner

MAX_ITEMS = 10000
OWNER_USER_ID = os.environ.get("BLUEPRINTS_USER_ID")
if not OWNER_USER_ID:
    print("Environment variable BLUEPRINTS_USER_ID is not set")
    exit(1)

user = User.objects.get(id=OWNER_USER_ID)
main_character = user.profile.main_character
character_ownership = CharacterOwnership.objects.get(
    user=user, character=main_character
)
try:
    corporation = EveCorporationInfo.objects.get(
        corporation_id=main_character.corporation_id
    )
except EveCorporationInfo.DoesNotExist:
    corporation = EveCorporationInfo.objects.create_corporation(
        main_character.corporation_id
    )

owner, _ = Owner.objects.update_or_create(
    corporation=corporation, character=character_ownership
)

# create blueprints for owner
eve_type, _ = EveType.objects.get_or_create_esi(id=692)  # stabber bp
location, _ = Location.objects.get_or_create_esi(
    id=60003760, token=owner.valid_token()
)  # jita 4-4

print(f"Generating {MAX_ITEMS} blueprints for {corporation}...")
blueprints = [
    Blueprint(
        item_id=randint(1_000_000_000_000, 1_999_999_999_999),
        owner=owner,
        eve_type=eve_type,
        location=location,
        location_flag="CorpDeliveries",
        quantity=1,
        runs=10,
        material_efficiency=10,
        time_efficiency=20,
    )
    for _ in range(MAX_ITEMS)
]
Blueprint.objects.bulk_create(blueprints, batch_size=500, ignore_conflicts=True)

print("DONE")



from typing import Tuple

from django.contrib.auth.models import User
from esi.models import Token

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import add_new_token

from blueprints.models import Owner


def create_owner(character_id, corporation_id) -> Owner:
    _, character_ownership = create_user_from_evecharacter(character_id)
    corp = (
        EveCorporationInfo.objects.get(corporation_id=corporation_id)
        if corporation_id
        else None
    )
    return Owner.objects.create(character=character_ownership, corporation=corp)


def create_user_from_evecharacter(character_id: int) -> Tuple[User, CharacterOwnership]:
    auth_character = EveCharacter.objects.get(character_id=character_id)
    user = AuthUtils.create_user(auth_character.character_name)
    user = AuthUtils.add_permission_to_user_by_name("blueprints.basic_access", user)
    user = AuthUtils.add_permission_to_user_by_name(
        "blueprints.add_corporate_blueprint_owner", user
    )
    user = AuthUtils.add_permission_to_user_by_name(
        "blueprints.add_personal_blueprint_owner", user
    )
    user = AuthUtils.add_permission_to_user_by_name("blueprints.manage_requests", user)
    character_ownership = add_character_to_user(
        user,
        auth_character,
        is_main=True,
        scopes=[
            "esi-assets.read_assets.v1",
            "esi-assets.read_corporation_assets.v1",
            "esi-characters.read_blueprints.v1",
            "esi-corporations.read_blueprints.v1",
            "esi-industry.read_character_jobs.v1",
            "esi-industry.read_corporation_jobs.v1",
            "esi-universe.read_structures.v1",
        ],
    )
    return user, character_ownership


def add_character_to_user(
    user: User,
    character: EveCharacter,
    is_main: bool = False,
    scopes: list = None,
) -> CharacterOwnership:
    if not scopes:
        scopes = "publicData"

    token = add_new_token(user, character, scopes)
    token.save()
    if is_main:
        user.profile.main_character = character
        user.profile.save()
        user.save()

    return CharacterOwnership.objects.get(user=user, character=character)


def scope_names_set(token: Token) -> set:
    return set(token.scopes.values_list("name", flat=True))



"""
Tools for building unit tests with django-esi
"""

from collections import namedtuple
from typing import Any, List

from bravado.exception import HTTPInternalServerError, HTTPNotFound

from django.utils.dateparse import parse_datetime


class BravadoResponseStub:
    """Stub for IncomingResponse in bravado, e.g. for HTTPError exceptions"""

    def __init__(
        self, status_code, reason="", text="", headers=None, raw_bytes=None
    ) -> None:
        self.reason = reason
        self.status_code = status_code
        self.text = text
        self.headers = headers if headers else dict()
        self.raw_bytes = raw_bytes

    def __str__(self):
        return "{0} {1}".format(self.status_code, self.reason)


class BravadoOperationStub:
    """Stub to simulate the operation object return from bravado via django-esi"""

    class RequestConfig:
        def __init__(self, also_return_response):
            self.also_return_response = also_return_response

    class ResponseStub:
        def __init__(self, headers):
            self.headers = headers

    def __init__(self, data, headers: dict = None, also_return_response: bool = False):
        self._data = data
        self._headers = headers if headers else {"x-pages": 1}
        self.request_config = BravadoOperationStub.RequestConfig(also_return_response)

    def result(self, **kwargs):
        if self.request_config.also_return_response:
            return [self._data, self.ResponseStub(self._headers)]
        else:
            return self._data

    def results(self, **kwargs):
        return self.result(**kwargs)


EsiEndpoint_T = namedtuple(
    "EsiEndpoint", ["category", "method", "primary_key", "needs_token"]
)


def EsiEndpoint(
    category: str, method: str, primary_key: str = None, needs_token: bool = False
) -> EsiEndpoint_T:
    return EsiEndpoint_T(category, method, primary_key, needs_token)


class _EsiRoute:
    def __init__(
        self, endpoint: EsiEndpoint_T, testdata: dict, http_error: bool = False
    ) -> None:
        self._category = endpoint.category
        self._method = endpoint.method
        self._primary_key = endpoint.primary_key
        self._needs_token = endpoint.needs_token
        self._testdata = testdata
        self._http_error = http_error

    def call(self, **kwargs):
        if self._http_error:
            raise HTTPInternalServerError(
                response=BravadoResponseStub(500, "Test exception")
            )

        pk_value = None
        if self._primary_key:
            if isinstance(self._primary_key, tuple):
                for pk in self._primary_key:
                    if pk not in kwargs:
                        raise ValueError(
                            f"{self._category}.{self._method}: Missing primary key: {pk}"
                        )

            elif self._primary_key not in kwargs:
                raise ValueError(
                    f"{self._category}.{self._method}: Missing primary key: "
                    f"{self._primary_key}"
                )
        if self._needs_token:
            if "token" not in kwargs:
                raise ValueError(
                    f"{self._category}.{self._method} "
                    f"with pk = {self._primary_key}: Missing token"
                )
            elif not isinstance(kwargs.get("token"), str):
                raise TypeError(
                    f"{self._category}.{self._method} "
                    f"with pk = {self._primary_key}: Token is not a string"
                )
        try:
            if self._primary_key:
                if isinstance(self._primary_key, tuple):
                    pk_value_1 = str(kwargs[self._primary_key[0]])
                    pk_value_2 = str(kwargs[self._primary_key[1]])
                    result = self._convert_values(
                        self._testdata[self._category][self._method][pk_value_1][
                            pk_value_2
                        ]
                    )
                else:
                    pk_value = str(kwargs[self._primary_key])
                    result = self._convert_values(
                        self._testdata[self._category][self._method][pk_value]
                    )
            else:
                result = self._convert_values(
                    self._testdata[self._category][self._method]
                )

        except KeyError:
            raise HTTPNotFound(
                response=BravadoResponseStub(
                    404,
                    f"{self._category}.{self._method}: No test data for "
                    f"{self._primary_key} = {pk_value}",
                ),
            ) from None

        return BravadoOperationStub(result)

    @staticmethod
    def _convert_values(data) -> Any:
        if isinstance(data, dict):
            for k, v in data.items():
                if isinstance(v, str):
                    try:
                        dt = parse_datetime(v)
                        if dt:
                            data[k] = dt
                    except ValueError:
                        pass

        return data


class EsiClientStub:
    def __init__(
        self, testdata: dict, endpoints: List[EsiEndpoint_T], http_error: bool = False
    ) -> None:
        self._testdata = testdata
        self._http_error = http_error
        for endpoint in endpoints:
            self._validate_endpoint(endpoint)
            self._add_endpoint(endpoint)

    def _validate_endpoint(self, endpoint: EsiEndpoint_T):
        try:
            _ = self._testdata[endpoint.category][endpoint.method]
        except KeyError:
            raise ValueError(f"No data provided for {endpoint}")

    def _add_endpoint(self, endpoint: EsiEndpoint):
        if not hasattr(self, endpoint.category):
            setattr(self, endpoint.category, type(endpoint.category, (object,), dict()))
        my_category = getattr(self, endpoint.category)
        if not hasattr(my_category, endpoint.method):
            setattr(
                my_category,
                endpoint.method,
                _EsiRoute(
                    endpoint=endpoint,
                    testdata=self._testdata,
                    http_error=self._http_error,
                ).call,
            )
        else:
            raise ValueError(f"Endpoint for {endpoint} already defined!")
